      INCLUDE 'flib.fi'
      
C$DEFINE SUBSIGNALQQ
$DEFINE SUBMATHERRQQ
      
$IF DEFINED (SUBSIGNALQQ)

      INTERFACE TO INTEGER*4 FUNCTION fpexhndl [C] ( signal, error )
      INTEGER*4 signal, error
      END
      
$ENDIF

c sadlfhskaldfhskaldf
CC    Illustrate floating point special values, 
CC    exceptions, and exception handling.
CC
CC    Uses: fpcontrl, fpexcept, fpexhndl, fpprecis, fpround, fpstatus,
CC          GETCONTROLFPQQ, GETSTATUSFPQQ, SETCONTROLFPQQ, SIGNALQQ
CC

C     The program initially exhibits the IEEE exception behavior.
C     All floating-point exceptions are disabled (all floating-point
C     exception flags are set) and the default MATHERRQQ run-time
C     subroutine is replaced with one in this file that does nothing
C     but return a successful error code.
C
C     For numerical precision most consistent with the IEEE standard,
C     you will additionally want to compile your programs with the
C     option /Op.  This example assumes that you compile it with /Op
C     to illustrate IEEE behavior.
C
C     You can try changing any of these settings to illustrate what
C     behavior results with those changes. If you want to illustrate
C     the default floating-point exception behavior (with resulting
C     run-time errors and program termination), you can remove one or
C     more flags from the call to "fpexcept".  You can also remove the
C     substitution of MATHERRQQ (by commenting out the definition of
C     SUBMATHERRQQ at the beginning of the file) to see the default
C     run-time math exception behavior.
C
C     To illustrate the use of a user-supplied signal handler, uncomment
C     the statement that defines SUBSIGNALQQ at the beginning of this file.
C     This activates "fpexhndl", a sample floating-point exception
C     handler.  "fpexhndl" is called only if you enable one or more
C     floating point exceptions (by not setting exception flags).
C  
      PROGRAM special

      INCLUDE 'flib.fd'

$IF DEFINED (SUBSIGNALQQ)

      EXTERNAL fpexhndl
      
$ENDIF

      INTEGER*2  FPCW$NULL
      PARAMETER (FPCW$NULL = #0000)

C     Variables
C
      INTEGER*2 status, control
      REAL u, v, w, x, y, z
      DOUBLE PRECISION du, dx, dy

C     Set up floating point exception handler
C
$IF DEFINED (SUBSIGNALQQ)

      INTEGER*4 signal
      
      signal = SIGNALQQ( SIG$FPE, fpexhndl )
      
$ENDIF

C     Determine initial control and status words
C
      CALL GETCONTROLFPQQ( control )
      CALL fpcontrl( control )
      WRITE(*,*)
      CALL GETSTATUSFPQQ( status )
      CALL FPSTATUS( status )
      WRITE(*,*)
      PAUSE 'Press any key to continue'

C     Check INVALID, DENORMAL, ZERODIVIDE, OVERFLOW, UNDERFLOW, INEXACT
C     Replace, e.g., FPCW$ZERODIVIDE with FPCW$NULL
C
      CALL fpexcept( FPCW$INVALID  + FPCW$DENORMAL  + FPCW$ZERODIVIDE +
     +               FPCW$OVERFLOW + FPCW$UNDERFLOW + FPCW$INEXACT )

C     Check 24, 53, 64
C     Replace 64 with 24 or 53
C
      CALL fpprecis( FPCW$64 )

C     Check NEAR, DOWN, UP, CHOP
C     Replace NEAR with DOWN, UP, or CHOP
C
      CALL fpround( FPCW$NEAR )
      
C     Determine current setting of control word
C
      CALL getcontrolfpqq( control )
      CALL fpcontrl( control )
      WRITE(*,*)

C     Check zerodivide (and signed infinity)
C
      x =  1.
      y = -1.
      z =  0.
      u = x / z
      CALL GETSTATUSFPQQ( status )
      CALL fpstatus( status )
      WRITE(*,*)
      v = y / z
      CALL GETSTATUSFPQQ( status )
      CALL fpstatus( status )
      WRITE(*,*)
      WRITE(*,*) '+ zerodivide = ', u, '   - zerodivide = ', v
      WRITE(*,'(18x,z,25x,z)') u, v
      WRITE(*,'(18x,e20.9,13x,e20.9)') u, v
      PAUSE 'Press any key to continue'

C     Check inexact (precision)
C
      x =  1.
      y =  0.00000001
      u = x + y
      CALL GETSTATUSFPQQ( status )
      CALL fpstatus( status )
      WRITE(*,*)
      v = x - y
      CALL GETSTATUSFPQQ( status )
      CALL fpstatus( status )
      WRITE(*,*)
      WRITE(*,*) '+ inexact = ', u, '   - inexact = ', v
      WRITE(*,'(20x,z,22x,z)') u, v
      WRITE(*,'(14x,e20.9,10x,e20.9)') u, v
      PAUSE 'Press any key to continue'

C     Check underflow
C
      x =  1.e-37
      y = -1.e-37
      z =  1.e+37
      u = x / z
      CALL GETSTATUSFPQQ( status )
      CALL fpstatus( status )
      WRITE(*,*)
      v = y / z
      CALL GETSTATUSFPQQ( status )
      CALL fpstatus( status )
      WRITE(*,*)
      WRITE(*,*) '+ underflow = ', u, '   - underflow = ', v
      WRITE(*,'(18x,z,24x,z)') u, v
      WRITE(*,'(12x,e20.9,12x,e20.9)') u, v
      PAUSE 'Press any key to continue'

C     Check overflow
C
      x =  1.e-37
      y = -1.e-37
      z =  1.e+37
      u = z / x
      CALL GETSTATUSFPQQ( status )
      CALL fpstatus( status )
      WRITE(*,*)
      v = z / y
      CALL GETSTATUSFPQQ( status )
      CALL fpstatus( status )
      WRITE(*,*)
      WRITE(*,*) '+ overflow = ', u, '   - overflow = ', v
      WRITE(*,'(16x,z,23x,z)') u, v
      WRITE(*,'(16x,e20.9,11x,e20.9)') u, v
      PAUSE 'Press any key to continue'

C     Check additional invalid operations
C
C     Check floating-point exception first
C
      x =  0.
      y =  0.
      z = -1.
      u = x / y
      CALL GETSTATUSFPQQ( status )
      CALL fpstatus( status )
      WRITE(*,*)
      WRITE(*,*) '+ 0/0 = ', u
      WRITE(*,'(11x,z)') u
      WRITE(*,'(11x,e20.9)') u
      PAUSE 'Press any key to continue'

C     Check run-time math exceptions second
C
      v = LOG( x )
      CALL GETSTATUSFPQQ( status )
      CALL fpstatus( status )
      WRITE(*,*)
      w = SQRT( z )
      CALL GETSTATUSFPQQ( status )
      CALL fpstatus( status )
      WRITE(*,*)
      WRITE(*,*) '+ log(0) = ', v, '   + sqrt(-1) = ', w
      WRITE(*,'(14x,z,23x,z)') v, w
      WRITE(*,'(14x,e20.9,11x,e20.9)') v, w
      WRITE(*,*)
      PAUSE 'Press any key to continue'

C     Check denormal
C
      dx = 1.0D-320
      dy =  0.1D0
      du = dx * dy
      CALL GETSTATUSFPQQ( status )
      CALL fpstatus( status )
      WRITE(*,*)
      WRITE(*,*) 'denorm = ', du
      WRITE(*,'(12x,z,21x,z)') du
      WRITE(*,'(11x,e25.19)') du
      WRITE(*,*)
      PAUSE 'Press any key to continue'

C     End with valid operation
C
      x = 1.
      y = 1.
      u = x / y
      CALL GETSTATUSFPQQ( status )
      CALL fpstatus( status )
      WRITE(*,*)

      END
      
$IF DEFINED (SUBMATHERRQQ)

CC    Replace the default run-time math error routine
CC    with one that does nothing. Therefore, run-time
CC    error messages do not occur and the program
CC    continues to run.

CC    NOTE: Because not all formal arguments declared are used in this
CC    version of MATHERRQQ, you will get a compiler warning that says
CC    "formal argument ... never used".
      
      SUBROUTINE MATHERRQQ(name, length, info, retcode)

      INCLUDE 'flib.fd'

      INTEGER*2 length
      CHARACTER name(length)
      RECORD / MTH$E_INFO / info
      INTEGER*2 retcode

      SELECT CASE (info.errcode)

         CASE (MTH$E_DOMAIN)

         CASE (MTH$E_SINGULARITY)

         CASE (MTH$E_OVERFLOW)

         CASE (MTH$E_UNDERFLOW)

         CASE (MTH$E_TLOSS)
 
         CASE (MTH$E_PLOSS)

         CASE DEFAULT

      END SELECT
      
      retcode = 1

      END
      
$ENDIF
      
