/* Updated 01/08/96 by Premia Corporation for Codewright v5.0 - S.Z.  */

/*͸*/
/*                                                                  */
/*  Program Name :  FOR.C                                           */
/*                                                                  */
/*        Author :  G. A. Szekely                                   */
/*                                                                  */
/*       Version :  1.0                                             */
/*                                                                  */
/* Creation Date :  08/04/94   06:49:16                             */
/*                                                                  */
/* Revision Date :  08/10/94                                        */
/*                                                                  */
/*͵*/
/*                                                                  */
/*   Description : This program is used with Premia Corporation's   */
/*                 Codewright vers 3.0g to add ChromaCoding and     */
/*                 Language Template features to a FORTRAN source   */
/*                 code file being edited using Codewright.         */
/*                 This FORTRAN version was created by modifying    */
/*                 the BASIC version that original came with        */
/*                 Codewright.                                      */
/*                                                                  */
/*;*/



#include <windows.h>
#include <string.h>
#include <ctype.h>
#include <exports.h>
#include "cwstart.h"

#define MAX_BUF_LEN   100

static LPVOID _for_hash_names = NIL;      /* pointer to the hash table containing */
                                          /* all FORTRAN keyword names. */

static LPVOID _for_hash_templates = NIL;  /* pointer to the hash table containing */
                                          /* an abbreviated name and its template */

static long _for_comment_col = 1;
static int _for_continue_col = 6;


/* Define a structure to hold an abbreviation and the template to be
 * expanded when the abbreviation is entered into the buffer.
 * This structure will be the element in the hash table _for_hash_templates.
 */
typedef struct _LanguageTemplate
   {
   LPSTR abbrev;
   LPSTR template;
   } LanguageTemplate;


/* This string contains all keyword names which will be indented an extra level.
 */
LPSTR FOR_INDENT_STR =
"\377IF\377ELSEIF\377ELSE\377DO\377PROGRAM\377SUBROUTINE\377FUNCTION\377STRUCTURE\377\
\377if\377elseif\377else\377do\377program\377subroutine\377function\377structure\377";
/*
 * _for_templates[] contains the keyword abbreviation and the template for the
 * abbreviation.  To add a new name, insert it before the NIL entry or
 * call _for_assign_template().
 *
 * The special characters within the template are:
 *
 *     \n - new line
 *     &  - cursor position after template insertion
 *     @  - backspace
 *     \c - insert 'c' literally
 */
LanguageTemplate _for_templates[] =
   {
      { ".F|.FA|.FAL|.FALS|.FALSE",  ".FALSE." },
      { ".T|.TR|.TRU|.TRUE",         ".TRUE." },
      { "B|BL|BLO|BLOC|BLOCK",       "BLOCK DATA%0&%1END" },
      { "CA|CAL|CALL",               "CALL & ()" },
      { "DA|DAT|DATA",               "DATA &" },
      { "DP|DOU|DOUB|DOUBL|DOUBLE",  "DOUBLE PRECISION" },
      { "D|DO",                      "DO I = 1, &%0%1ENDDO" },
      { "EI",                        "ELSE IF (&)%1" },
      { "EL|ELS|ELSE",               "ELSE%1" },
      { "EN|END",                    "END" },
      { "FO|FOR|FORM|FORMA|FORMAT",  "FORMAT (&)" },
      { "G|GO",                      "GO TO &" },
      { "IF|IF",                     "IF (&) THEN%0%1ENDIF" },
      { "M|MA|MAP",                  "MAP%0&%1END MAP" },
      { "REAL",                      "REAL*8 &" },
      { "REC|RECO|RECOR|RECORD",     "RECORD &" },
      { "RET|RETU|RETUR|RETURN",     "RETURN" },
      { "REW|REWI|REWIN|REWIND",     "REWIND &" },
      { "R|RE|REA|READ",             "READ (5, &)" },
      { "S|ST|STO|STOP",             "STOP" },
      { "U|UN|UNI|UNIO|UNION",       "UNION%0&%1END UNION" },
      { "W|WR|WRI|WRIT|WRITE",       "WRITE (6, &)" },
      { "CH|CHA|CHAR|CHARA|CHARAC|CHARACT|CHARACTE|CHARACTER", "CHARACTER*4 &" },
      { "CO|CON|CONT|CONTI|CONTIN|CONTINU|CONTINUE",           "CONTINUE" },
      { "DI|DIM|DIME|DIMEN|DIMENS|DIMENSI|DIMENSIO|DIMENSION", "DIMENSION &()" },
      { "F|FU|FUN|FUNC|FUNCT|FUNCTI|FUNCTIO|FUNCTION",         "FUNCTION & ()%0%1END" },
      { "IM|IMP|IMPL|IMPLI|IMPLIC|IMPLICI|IMPLICIT",           "IMPLICIT REAL*8 (A-H, O-Z)" },
      { "IN|INT|INTE|INTEG|INTEGE|INTEGER",                    "INTEGER*2 &" },
      { "LO|LOG|LOGI|LOGIC|LOGICA|LOGICAL",                    "LOGICAL*1 &" },
      { "P|PR|PRO|PROG|PROGR|PROGRA|PROGRAM",                  "PROGRAM & ()%0%1END" },
      { "ST|STR|STRU|STRUC|STRUCT|STRUCTU|STRUCTUR|STRUCTURE", "STRUCTURE & %0%1END" },
      { "S|SU|SUB|SUBR|SUBRO|SUBROU|SUBROUT|SUBROUTINE",       "SUBROUTINE & ()%0%1END" },

      { ".f|.fa|.fal|.fals|.false",  ".false." },
      { ".t|.tr|.tru|.true",         ".true." },
      { "b|bl|blo|bloc|block",       "block data%0&%1end" },
      { "ca|cal|call",               "call & ()" },
      { "da|dat|data",               "data &" },
      { "dp|dou|doub|doubl|double",  "double precision" },
      { "d|do",                      "do i = 1, &%0%1enddo" },
      { "ei",                        "else if (&)%1" },
      { "el|els|else",               "else%1" },
      { "en|end",                    "end" },
      { "fo|for|form|forma|format",  "format (&)" },
      { "g|go",                      "go to &" },
      { "if|if",                     "if (&) then%0%1endif" },
      { "m|ma|map",                  "map%0&%1end map" },
      { "real",                      "real*8 &" },
      { "rec|reco|recor|record",     "record &" },
      { "ret|retu|retur|return",     "return" },
      { "rew|rewi|rewin|rewind",     "rewind &" },
      { "r|re|rea|read",             "read (5, &)" },
      { "s|st|sto|stop",             "stop" },
      { "u|un|uni|unio|union",       "union%0&%1end union" },
      { "w|wr|wri|writ|write",       "write (6, &)" },
      { "ch|cha|char|chara|charac|charact|characte|character", "character*4 &" },
      { "co|con|cont|conti|contin|continu|continue",           "continue" },
      { "di|dim|dime|dimen|dimens|dimensi|dimensio|dimension", "dimension &()" },
      { "f|fu|fun|func|funct|functi|functio|function",         "function & ()%0%1end" },
      { "im|imp|impl|impli|implic|implici|implicit",           "implicit real*8 (a-h, o-z)" },
      { "in|int|inte|integ|intege|integer",                    "integer*2 &" },
      { "lo|log|logi|logic|logica|logical",                    "logical*1 &" },
      { "p|pr|pro|prog|progr|progra|program",                  "program & ()%0%1end" },
      { "st|str|stru|struc|struct|structu|structur|structure", "structure & %0%1end" },
      { "s|su|sub|subr|subro|subrou|subrout|subroutine",       "subroutine & ()%0%1end" },

      { NIL,     NIL },
   };

/* Define the list of FORTRAN keyword names to appear in the keyword color
   if language coloring is enabled.
 */
LPSTR _for_names[] =
   {
   /*  FORTRAN Logical Operators  */
   "AND",          "EQ",           "EQV",          "FALSE",        "GE",
   "GT",           "LE",           "LT",           "NE",           "NEQV",
   "NOT",          "OR",           "TRUE",         "XOR",

   /*  FORTRAN Keywords  */
   "ACCESS",       "ALIAS",        "ALLOCATABLE",  "ALLOCATE",     "ALLOCATED",
   "APPEND",       "ASSIGN",       "AUTOMATIC",    "BACKSPACE",    "BINARY",
   "BLANK",        "BLOCK",        "BLOCKSIZE",    "BYTE",         "CALL",
   "CASE",         "CHARACTER",    "CLOSE",        "COMMON",       "COMPLEX",
   "CONTINUE",     "CYCLE",        "DATA",         "DEALLOCATE",   "DECODE",
   "DIMENSION",    "DO WHILE",     "DO",           "DOUBLE",       "ELSE",
   "ELSEIF",       "ENCODE",       "END",          "ENDDO",        "ENDIF",
   "ENDFILE",      "ENTRY",        "EQUIVALENCE",  "ERR",          "EXIT",
   "EXTERN",       "EXTERNAL",     "FAR",          "FILE",         "FMT",
   "FOR",          "FORM",         "FORMAT",       "FORMATTED",    "FUNCTION",
   "GO",           "GOTO",         "IF",           "I/O",          "IMPLICIT",
   "INCLUDE",      "INQUIRE",      "INQUIRE",      "INTEGER",      "INTERFACE",
   "INTRINSIC",    "IOSTAT",       "KEEP",         "LINE",         "LOCKING",
   "LOCKMODE",     "LOGICAL",      "MAP",          "MODE",         "NAMED",
   "NAMELIST",     "NEAR",         "NEW",          "NEXTREC",      "NML",
   "OLD",          "OPEN",         "OPENED",       "PARAMETER",    "PAUSE",
   "PRECISION",    "PRINT",        "PROGRAM",      "READ",         "REAL",
   "REC",          "RECL",         "RECORD",       "RECORDS",      "REFERENCE",
   "RETURN",       "REWIND",       "SAVE",         "SELECT",       "SEQUENTIAL",
   "SHARE",        "STAT",         "STATUS",       "STOP",         "STRUCTURE",
   "SUBROUTINE",   "THEN",         "TO",           "UNFORMATTED",  "UNION",
   "UNIT",         "VALUE",        "WRITE",

   /*  FORTRAN Intrinsic Functions  */
   "ABS",          "ACOS",         "AIMAG",        "AINT",         "ALLOCATED",
   "ALOG",         "ALOG10",       "AMAX0",        "AMAX1",        "AMIN0",
   "AMIN1",        "AMOD",         "ANINT",        "ASIN",         "ATAN",
   "ATAN2",        "BTEST",        "CABS",         "CCOS",         "CDABS",
   "CDCOS",        "CDEXP",        "CDLOG",        "CDSIN",        "CDSQRT",
   "CEXP",         "CHAR",         "CLOG",         "CMPLX",        "CONJG",
   "COS",          "COSH",         "COTAN",        "CSIN",         "CSQRT",
   "DABS",         "DACOS",        "DASIN",        "DATAN",        "DATAN2",
   "DBLE",         "DCMPLX",       "DCONJG",       "DCOS",         "DCOSH",
   "DCOTAN",       "DDIM",         "DEXP",         "DFLOAT",       "DIM",
   "DIMAG",        "DININT",       "DINT",         "DLOG",         "DLOG10",
   "DMAX1",        "DMIN1",        "DMOD",         "DNINT",        "DPROD",
   "DREAL",        "DSIGN",        "DSIN",         "DSINH",        "DSQRT",
   "DTAN",         "DTANH",        "EOF",          "EPSILON",      "EXP",
   "FLOAT",        "HFIX",         "HUGE",         "IABS",         "IAND",
   "IBCHNG",       "IBCLR",        "IBSET",        "ICHAR",        "IDIM",
   "IDINT",        "IDNINT",       "IEOR",         "IFIX",         "IMAG",
   "INDEX",        "INT",          "INT1",         "INT2",         "INT4",
   "INTC",         "IOR",          "ISHA",         "ISHC",         "ISHFT",
   "ISHL",         "ISIGN",        "JFIX",         "LEN",          "LEN_TRIM",
   "LGE",          "LGT",          "LLE",          "LLT",          "LOC",
   "LOCFAR",       "LOCNEAR",      "LOG",          "LOG10",        "MAX",
   "MAX0",         "MAX1",         "MAXEXPONENT",  "MIN",          "MIN0",
   "MIN1",         "MINEXPONENT",  "MOD",          "NEAREST",      "NINT",
   "SCAN",         "SETDAT",       "SIGN",         "SIN",          "SINH",
   "SNGL",         "SQRT",         "TAN",          "TANH",         "TINY",
   "VERIFY",

   NIL
   };



static long _for_in_comment (long offset);

long DLL _for_comment_column (long col);
void DLL _init (void);
BOOL DLL _for_expand (void);
long DLL _for_language_colors_update (long firstline, long lastline);
void DLL _for_language_colors (long numlines);
void DLL _for_assign_template (LPSTR abbrev, LPSTR template);
void DLL _for_add_keyword (LPSTR keyword);
BOOL DLL _for_indent (void);


/******************************************************************************
 * This section contains sample source code to be used with the Microsoft 'C' *
 * compiler.  No SDK is required unless a resource file is to be used.        *
 * The functions in the SDK can be accessed by linking with a library         *
 * created by running ilib.exe on the USER.EXE and GDI.EXE in your            *
 * Window's System directory.                                                 *
 ******************************************************************************/

LIBMAIN            /* defines the LibMain () function for Borland C */


/* When this DLL is first loaded either by LibPreLoad() or LibAutoLoad(),
 * the function _init () is invoked.  Within this function, all DLL specific
 * initialization should be performed such as:
 *
 *         - Defining any module global keystrokes or keymaps to be used
 *         - Exporting any functions to be accessed through LibFunctionExec()
 *         - Autoloading any other DLLs.
 *
 */

void DLL _init ()
   {
   /*  Place all 'LibExport' calls here.            */
   /*  i.e.  LibExport ("int myfunc LPSTR DWORD");  */

   static BOOL firstTime = TRUE;
   int i;

   if (firstTime)
      {
      LibExport ("BOOL _for_expand");
      LibExport ("_for_assign_template LPSTR LPSTR");
      LibExport ("_for_add_keyword LPSTR");
      LibExport ("BOOL _for_indent");
      LibExport ("long _for_language_colors_update long long");
      LibExport ("_for_language_colors long");

      firstTime = FALSE;

      /* Create a hash table to hold all of the FORTRAN keyword names.  */
      _for_hash_names = HashCreateTable (59, /* a prime # */
                                         sizeof (LPSTR), 0);
      i = 0;

      while (_for_names[i] != NIL)
         {
         LPSTR XFAR *lpf = (LPSTR XFAR *) HashGetEntry (_for_hash_names,
                                                        _for_names[i], strlen (
                                                        _for_names[i]));

         if (lpf)
            {
            *lpf = StrNew (_for_names[i]);
            i++;
            }   /* if lpf */
         }   /* while _for_names */

      /* Create a hash table to hold all of the FORTRAN template names.  */
      _for_hash_templates = HashCreateTable (61, /* a prime # */
                                             sizeof (LPSTR), 0);
      i = 0;

      while (_for_templates[i].abbrev != NIL)
         {
         _for_assign_template (_for_templates[i].abbrev, _for_templates[i].
                               template);
         i++;
         }   /* while _for_templates */

      CompilerNewExt (".for");
      }   /* if firstTime */
   }   /* function DLL _init */


PUBLIC long DLL _for_comment_column (long col)
   {
   if (col > 0)
      _for_comment_col = col;
   return _for_comment_col;
   }   /* function DLL _for_comment_column */


/*
 *  _for_expand ()
 *
 *  DESCRIPTION:
 *    Attempts to perform FORTRAN language template expansion.  This function
 *    is normally bound to the <SPACE> key.  When invoked, it reads the
 *    previous word.  If the word is in the list of defined templates,
 *    the template is extracted and inserted.  Some characters in the
 *    templates have special meanings. They are:
 *
 *       \n - new line
 *       &  - cursor position after template insertion
 *       @  - backspace
 *       \  - take next character literally
 */
BOOL DLL _for_expand (void)
   {
   LPVOID htEntry;
   LPSTR _for_string = NULL;
   LPSTR str;

   MarkSavePos ();

   /* Check to see if we are currently positioned inside a FORTRAN comment
    */
   if (_for_in_comment (-1L) != -1)
      goto _for_expand_incomplete;

   str = BufReadStr (0x7FFFFFFFL);     /* read entire line */

   _for_string = StrTrim (str, " \t");
   StrFree (str);

   if (!_for_string)
      goto _for_expand_incomplete;

   /* Search for a keyword match in the hash table.
    */
   htEntry = HashFindEntry (_for_hash_templates, _for_string, strlen (
                            _for_string));
   StrFree (_for_string);

   if (htEntry)
      {
      /* A template exists for the current keyword, delete the keyword
       * and insert the template.
       */
      MarkDropPos (1);
      str = *(LPSTR *) htEntry;
      BufDelToEOL ();

      _ext_expand_template (str);
      }   /* if htEntry */
   else
      goto _for_expand_incomplete;
   return TRUE;

_for_expand_incomplete:
   MarkRestorePos ();

   return FALSE;
   }   /* function DLL _for_expand */


/*
 *  _for_language_colors_update ()
 *
 *  This function does an intermediate update of the colors within
 *  a range of line numbers (firstline and lastline).  It must also
 *  determine if the first and last line specified are within a
 *  comment and act appropriately.
 */
long DLL _for_language_colors_update (long firstline, long lastline)
   {
   long startoff;
   long count;

   if (firstline && lastline)
      {
      /* Begin coloring at the previous comment if
       * currently located in the middle of one.
       */
      _PosInit (0);
      if (firstline == 1)
         _PosPrevLine (0L);
      else
         _PosNextLine (firstline - 1);
      startoff = _PosQOffset ();

      /* Clear all attributes for the line range and
       * the reset them within _for_language_colors.
       */
      AttrSetColor (firstline, lastline, 1L, 0x7FFFFFFFL, 0);
      _PosInit (startoff);
      _for_language_colors (lastline - firstline + 1);
      }   /* if firstline */

      count = lastline - firstline;
      return(count);
   }   /* function DLL _for_language_colors_update */


/*
 **  _for_language_colors ()
 *
 *  If numlines == -1, start from the top of the buffer and encode all of them,
 *  otherwise start at the current location of the number of lines specified.
 *
 *  It is assumed that if numlines != -1, that _PosInit() has been called
 *  and numlines is relative to the current position.
 */
void DLL _for_language_colors (long numlines)
   {
   long startoff;
   long startline;
   long endoff;
   long inc;
   WORD ch;
   LPSTR buf;
   int beginline = 1;
   int columnnum = 1;
   int i = 0;
   int comment_color = ColorComments (-1);
   int keyword_color = ColorKeywords (-1);
   long maxlines = ExtCommentSearchLimit (-1);
   int quote_color = ColorAlternate1 (-1);
   int number_color = ColorAlternate2 (-1);
   int brace_color  = ColorAlternate3( -1 );
   int preproc_color = ColorAlternate4 (-1);


   /* Set Visible Left Margin at column number 72                    */
   WinVisibleMarginColumn (72);


   buf = (LPSTR) MemAlloc (MAX_BUF_LEN);

   /* Start at the top of the buffer or current position
    */
   if (numlines == -1)
      {
      _PosInit (0);
      numlines = 0x7FFFFFFFL;
      }   /* if numlines */
   else
      {
      startoff = _PosQOffset ();
      startline = _PosQLine ();
      endoff = _for_in_comment (startoff);
      if (endoff == -1)
         _PosInit (startoff);
      else
         {
         _PosInit (endoff);
         numlines += (startline - _PosQLine ());
         }   /* else endoff */
      }   /* else numlines */
   ch = _PosCurrentChar ();
   startoff = _PosQOffset ();

   maxlines = max (numlines, maxlines);

   while ((ch != EOF_CHAR) && (numlines > 0))
      {
      if (ch == '\n')
         {
         numlines--;
         beginline = 1;
         columnnum = 1;
         }   /* if ch */

      if (ch == ' ')
         {
         beginline = 0;
         columnnum++;
         }   /* if ch */

      /*  C or c in column 1 Comments  */
      if (((ch == 'C') || (ch == 'c')) && beginline)
         {
         startoff = _PosQOffset ();

         inc = _PosNextLine (1);
         numlines--;
         beginline = 1;
         columnnum = 1;
         endoff = _PosQOffset ();
         _PosInit (startoff);
         _PosSetColor (comment_color, endoff - startoff - inc);
         _PosInit (endoff);
         ch = _PosCurrentChar ();
         continue;
         }   /* if ch */

      /*  Quoted  */
      if (ch == '\'')
         {
         UINT quote = ch;

         beginline = 0;

         startoff = _PosQOffset ();

         do {
            ch = _PosNextChar ();
            if (ch == '\\')
               {
               ch = _PosNextChar ();
               if ((ch == quote) || (ch == '\\'))
                  ch = _PosNextChar ();
               }   /* if ch */
            if ((ch == EOF_CHAR) || (ch == '\n'))
               {
               beginline = 1;
               columnnum = 1;
               break;
               }   /* if ch */
            } while (ch != quote);

         if (ch == quote)
            ch = _PosNextChar ();

         /* color the string */
         if (quote_color)
            {
            endoff = _PosQOffset ();
            _PosInit (startoff);
            _PosSetColor (quote_color, endoff - startoff);
            _PosInit (endoff);
            }   /* if quote_color */
         continue;
         }   /* if ch */

      /* parenthesis color */
      else if (brace_color && _fstrchr("()", ch) != 0)
         {
         long coff;

         coff = _PosQOffset ();
         _PosInit (coff);
         _PosSetColor (brace_color, 1);
         _PosInit (coff);
         }  /* else if brace_color */

       /* preprocessor commands  */
       else if ( ch == '$' && preproc_color && (columnnum == 1))
         {
         startoff = _PosQOffset ();
         do {
            ch = _PosNextChar ();
            } while (ch == ' ' || ch == '\t');
         do {
            ch = _PosNextChar();
            } while ( !isspace(ch) && ch != EOF_CHAR);
         endoff = _PosQOffset();
         _PosInit (startoff);
         _PosSetColor (preproc_color, endoff - startoff);
         _PosInit (endoff);
         continue;
         }   /* else if ch */

      /*  Digit  */
      else if (number_color && (columnnum != _for_continue_col) && isdigit( (
               BYTE)ch ))
         {
         int j = 0;
         UINT lastch = _PosPrevChar ();

         beginline = 0;

         if (lastch != EOF_CHAR)
            _PosNextChar ();

         startoff = _PosQOffset ();
         do {
            j++;
            ch = _PosNextChar ();
            columnnum++;
            } while (isxdigit( (BYTE)ch ) || (ch == '.') || (ch == 'x' && j ==
                     1));

         if ((lastch != ' '
             && lastch != '\t'
             && lastch != '\r'
             && lastch != '\n'
             && lastch != EOF_CHAR
             && !ispunct( (BYTE)lastch ))
             || lastch == '_'
             || (ch == '_'))
            continue;

         if (ch == 'l' || ch == 'L')
            ch = _PosNextChar ();

         endoff = _PosQOffset ();
         _PosInit (startoff);
         _PosSetColor (number_color, endoff - startoff);
         _PosInit (endoff);
         continue;
         }   /* else if number_color */

      /*  Alpha  */
      else if ((isalpha( (BYTE)ch ) || (ch == '_')) && (columnnum !=
               _for_continue_col))
         {
         UINT lastch = _PosPrevChar ();

         beginline = 0;

         if (lastch != EOF_CHAR)
            _PosNextChar ();

         i = 0;
         startoff = _PosQOffset ();

         do {
            if (i < MAX_BUF_LEN)
               buf[i++] = (char) ch;
            ch = _PosNextChar ();
            columnnum++;
            } while (isalpha( (BYTE)ch ) || (ch == '_'));

         if ((lastch != ' '
             && lastch != '\t'
             && lastch != '\r'
             && lastch != '\n'
             && lastch != EOF_CHAR
             && !ispunct( (BYTE)lastch ))
             || (ch == '_'))
            i = 0;
         continue;
         }   /* else if isalpha */
      else if (i)
         {
         beginline = 0;
         if (HashFindEntry (_for_hash_names, _fstrupr(buf), i))
            {
            endoff = _PosQOffset ();
            _PosInit (startoff);
            _PosSetColor (keyword_color, endoff - startoff);
            _PosInit (endoff);
            }   /* if HashFindEntry */
         i = 0;
         }   /* else if i */

      /*  ! Comments  */
      if (ch == '!')
         {
         startoff = _PosQOffset ();

         inc = _PosNextLine (1);
         numlines--;
         beginline = 1;
         columnnum = 1;
         endoff = _PosQOffset ();
         _PosInit (startoff);
         _PosSetColor (comment_color, endoff - startoff - inc);
         _PosInit (endoff);
         ch = _PosCurrentChar ();
         continue;
         }   /* if ch */

      ch = _PosNextChar ();
      }   /* while ch */

   MemFree (buf);
   }   /* function DLL _for_language_colors */


void DLL _for_add_keyword (LPSTR keyword)
   {
   if (keyword && *keyword)
      {
      LPSTR XFAR *lpf = (LPSTR XFAR *) HashGetEntry (_for_hash_names,
                                                     _fstrupr(keyword),
                                                     strlen (keyword));

      if (lpf)
         *lpf = StrNew (keyword);
      }   /* if keyword */
   }   /* function DLL _for_add_keyword */


/*
 ** _for_assign_template ()
 *
 *  PARAMETERS:
 *    abbrev   - the keyword abreviation used to locate a template
 *    template - the encoded template string associated with 'abbrev'
 *
 *  DESCRIPTION:
 *    Provide a means of adding or modifying FORTRAN template strings
 *    without changing the code.
 */
void DLL _for_assign_template (LPSTR abbrev, LPSTR template)
   {
   _ext_assign_template (_for_hash_templates, abbrev, template);
   }   /* function DLL _for_assign_template */


/*
 ** _for_indent ()
 *
 *  DESCRIPTION:
 *    Attempts to perform FORTRAN language smart indenting.  This function
 *    will behave differently fored on the following conditions:
 *
 *    1. Within a comment - normal indentation
 *
 *    2. Not positioned at end of line - normal indentation
 *
 *    3. Current line prefixed by a name in FOR_INDENT_STR - perform smart
 *       indentation.  This may include skipping the next line if
 *       it consists only of an open brace.
 */
BOOL DLL _for_indent (void)
   {
   long curline;
   long col;
   LPSTR _for_string;
   LPSTR str;
   long matchpos;
   WORD endch;

   MarkSavePos ();

   curline = BufQCurrentLine ();

   /* Check to see if we are currently positioned inside a FORTRAN comment
    */
   if (_for_in_comment (-1L) != -1)
      goto _for_indent_incomplete;

   /* Check to see if there is a trailing '{' on the current line,
    * and if so, indent on the next line.
    */
   do {
      MovPrevChar (1L);
      endch = BufReadChar ();
      } while (endch == ' ' || endch == '\t');
   MarkRestorePos ();
   if (endch == '{')
      goto _for_indent_next_line;
   MarkSavePos ();

   /* Not in a comment so indent according to the current FORTRAN construct.
    */
   if (!SrchFind ("^[ \t]*[a-zA-Z_]+\\c([ \t]|$)", SEARCH_REGEX, NIL)
       || (BufQCurrentLine () != curline))
      goto _for_indent_incomplete;

   /* Read the keyword at the beginning of the current line.
    */
   col = BufQCurrentCol ();
   SrchFind ("(^|[^a-zA-Z_])\\c[a-zA-Z_]", SEARCH_REGEX, NIL);
   str = BufReadStr (col - BufQCurrentCol ());
   _for_string = StrTrim (str, " \t");
   StrFree (str);

   /* Set prefix and suffix to \377 to perform a quick lookup of the name
    */
   str = StrNew ("\377");
   _for_string = StrAppend (_for_string, str);
   str = StrAppend (str, _for_string);
   StrFree (_for_string);

   matchpos = StrMatch (str, FOR_INDENT_STR, SEARCH_FORWARD, NIL);
   StrFree (str);

   if (!matchpos)
      /* Not a FORTRAN keyword, do normal indentation. */
      goto _for_indent_incomplete;

   MarkRestorePos ();
   MarkSavePos ();

_for_indent_next_line:
   _ext_insert_indented_EOL ();
   BufInsertChar ('\t');
   return TRUE;

_for_indent_incomplete:
   MarkRestorePos ();

   return FALSE;
   }   /* function DLL _for_indent */




/*--  Private functions --*/


/*
 ** _for_in_comment ()
 *
 *  Determines whether or not the current position in the buffer is
 *  located within a FORTRAN comment by search backward for a !
 *  comment declaration.
 *
 *  Returns -1 if not in a comment
 *          >= 0 indicates the offset in the buffer where the comment begins
 */
static long _for_in_comment (long offset)
   {
   WORD ch, chk;

   /*  Check to see if we are currently positioned inside a 'FORTRAN' comment.
    *  by searching backwards for a 'FORTRAN' line comment symbol "!", and
    *  by searching backwards for a traditional "FORTRAN" comment symbol of
    *  a C in column 1.
    */

   _PosInit (offset);
   ch = _PosCurrentChar ();

   do {
      if (ch == '\n')
         return -1L;
      else if (ch == '!')
         return _PosQOffset ();  /* currrent line is in a "!" style comment */
      else if ((ch == 'C') || (ch == 'c'))
         {
         chk = _PosPrevChar ();  /* is the C is the first character line    */
         if ((chk == '\n') || (chk == EOF_CHAR))
            return _PosQOffset ();  /* currrent line is in a "C" style comment */
		 if (chk != EOF_CHAR)
		    _PosNextChar();
         }

      ch = _PosPrevChar ();
      } while (ch != EOF_CHAR);

   return -1L;
   }   /* function _for_in_comment */


